const els = {};
const vis = {};
let game = {};
const predef_games = {};

vis.canvas_skill_width = 250;
vis.canvas_skill_height = 250;
vis.canvas_player_width = 300;
vis.canvas_player_height = 300;
vis.canvas_score_width = 800;
vis.canvas_score_height = 300;
vis.canvas_player_skill_margin = 0.1;
vis.next_countdown_init = 4;
vis.colors_players = ["blue", "red", "darkgreen", "darkorange", "indigo", "lime"];
vis.colors_skills = ["#ffcd96", "#8fffc7", "#ffd1e7", "#8990ff", "#ff95bb", "#cbb5ff", "#d3ff93", "#ff8e85", "#bdccff", "#fff09b", "#ff9298", "#ff82a0"];

function init() {
    els.log = document.getElementById('log');
    els.jsloader = document.getElementById('jsloader');

    els.vis_c = document.getElementById('vis_c');
    els.vis_canvas_skills = document.getElementById('vis_canvas_skills');
    els.vis_canvas_players = document.getElementById('vis_canvas_players');
    els.vis_canvas_score = document.getElementById('vis_canvas_score');
    els.vis_evalnum = document.getElementById('vis_evalnum');
    els.vis_roundnum = document.getElementById('vis_roundnum');

    els.nav_c = document.getElementById('nav_c');
    els.nav_toggle = document.getElementById('nav_toggle');
    els.nav_timeout = document.getElementById('nav_timeout');
    els.nav_prev = document.getElementById('nav_prev');
    els.nav_next = document.getElementById('nav_next');
    els.nav_eval = document.getElementById('nav_eval');
    els.nav_round = document.getElementById('nav_round');

    document.onkeydown = function (evt) {
        evt = evt || window.event;
        if (evt.keyCode === 33 || evt.keyCode === 37) {
            // page up || left arrow
            vis_prev();
            evt.preventDefault();
        } else if (evt.keyCode === 34 || evt.keyCode === 39) {
            // page down || right arrow
            vis_next();
            evt.preventDefault();
        } else if (evt.keyCode === 190 || evt.keyCode === 32) {
            // period || space
            vis_toggle();
            evt.preventDefault();
        }
    };
}

function parseInt10(x) {
    return parseInt(x);
}

function vis_clear() {
    let fc;
    while (fc = els.vis_canvas_skills.firstChild) {
        els.vis_canvas_skills.removeChild(fc);
    }
    while (fc = els.vis_canvas_players.firstChild) {
        els.vis_canvas_players.removeChild(fc);
    }
    while (fc = els.vis_canvas_score.firstChild) {
        els.vis_canvas_score.removeChild(fc);
    }
}

function vis_init() {
    vis_clear();
    vis.next_countdown = 0;
    vis.current_eval = 0;
    vis.current_round = 0;
    vis.playing = false;
    vis.animation = -1;
    vis.timeout = parseInt(els.nav_timeout.value);
    vis.max_skill = Math.max.apply(null, game.skills);
    vis.max_score = Math.max.apply(null, game.endscore);
    els.nav_eval.max = game.evals.length;
    vis.total_rounds = 0;
    for(let i = 0; i < game.evals.length; i++) {
        vis.total_rounds += game.evals[i].length;
    }

    els.vis_skills = [];
    for (let i = 0; i < game.skillcount; i++) {
        const div = document.createElement('div');
        const canv = document.createElement('canvas');
        div.appendChild(canv);
        els.vis_canvas_skills.appendChild(div);
        div.classList.add('floatingdiv');

        canv.id = `vis_canvas_skill_${i}`;
        canv.width = vis.canvas_skill_width;
        canv.height = vis.canvas_skill_height;
        els.vis_skills.push(canv);
    }

    els.vis_players = [];
    els.vis_player_scores_eval = [];
    els.vis_player_scores_total = [];
    for (let i = 0; i < game.playercount; i++) {
        const div = document.createElement('div');
        const canv = document.createElement('canvas');
        const p1 = document.createElement('p');
        const p2 = document.createElement('p');
        const name = document.createElement('b');
        const score_eval = document.createElement('span');
        const score_total = document.createElement('span');
        name.appendChild(document.createTextNode(game.players[game.permutation[i]]));
        score_eval.appendChild(document.createTextNode('0'));
        score_total.appendChild(document.createTextNode('0'));
        p1.appendChild(name);
        p1.appendChild(document.createTextNode(': '));
        p1.appendChild(score_eval);
        p1.appendChild(document.createTextNode(' / '));
        p1.appendChild(score_total);
        p2.appendChild(canv);
        div.appendChild(p1);
        div.appendChild(p2);
        els.vis_canvas_players.appendChild(div);
        div.classList.add('floatingdiv');
        els.vis_player_scores_eval.push(score_eval);
        els.vis_player_scores_total.push(score_total);

        canv.id = `vis_canvas_player_${i}`;
        canv.width = vis.canvas_player_width;
        canv.height = vis.canvas_player_height;
        els.vis_players.push(canv);
    }

    const vis_score = document.createElement('canvas');
    els.vis_score = vis_score;
    vis_score.id = `vis_canvas_score_0`;
    vis_score.width = vis.canvas_score_width;
    vis_score.height = vis.canvas_score_height;
    els.vis_canvas_score.appendChild(vis_score);

    vis_apply_state();
}

function vis_apply_state() {
    els.nav_eval.value = vis.current_eval + 1;
    els.nav_round.value = vis.current_round + 1;
    els.nav_round.max = game.evals[vis.current_eval].length;
    els.vis_evalnum.innerHTML = vis.current_eval + 1;
    els.vis_roundnum.innerHTML = vis.current_round + 1;

    const ev = game.evals[vis.current_eval];
    const cev = ev[vis.current_round];
    const sc = game.skillcount;
    const pc = game.playercount;

    // skills
    const soh = vis.canvas_skill_height / (vis.max_skill + 2);
    const sow = vis.canvas_skill_width / ev.length;
    const scs = [];
    for (let as = 0; as < sc; as++) {
        const ctx = els.vis_skills[as].getContext('2d');
        if (cev.skillstatus[as] === 0) {
            ctx.fillStyle = '#eeeeee';
        } else if (cev.skillstatus[as] === 1) {
            ctx.fillStyle = '#ffe18b';
        } else {
            ctx.fillStyle = '#848484';
        }
        ctx.fillRect(0, 0, ctx.canvas.width, ctx.canvas.height);
        scs.push(ctx);
    }

    for (let as = 0; as < sc; as++) {
        const ctx = scs[as];
        ctx.fillStyle = vis.colors_skills[as % vis.colors_skills.length];
        ctx.beginPath();
        ctx.fillRect(0, 0, vis.canvas_skill_width, 0.5 * soh);
        ctx.strokeStyle = '#000000';
        ctx.lineWidth = 1;
        ctx.beginPath();
        const y = vis.canvas_skill_height - (game.skills[as] + 1) * soh;
        ctx.moveTo(0, y);
        ctx.lineTo(vis.canvas_skill_width, y);
        ctx.stroke();
        for (let ap = 0; ap < pc; ap++) {
            ctx.strokeStyle = vis.colors_players[ap % vis.colors_players.length];
            ctx.lineWidth = 2;
            ctx.beginPath();
            ctx.moveTo(0, vis.canvas_skill_height - soh);
            for (let ar = 0; ar <= vis.current_round; ar++) {
                ctx.lineTo((ar + 1) * sow, vis.canvas_skill_height - (ev[ar].skills[ap][as] + 1) * soh);
            }
            ctx.stroke();
        }
    }

    // players
    const poh = vis.canvas_player_height / (vis.max_skill + 1);
    const pow = vis.canvas_player_width / sc;
    const pcs = [];

    for (let ap = 0; ap < pc; ap++) {
        const ctx = els.vis_players[ap].getContext('2d');
        if (cev.target_player === ap) {
            ctx.fillStyle = '#ffe18b';
        } else {
            ctx.fillStyle = '#eeeeee';
        }
        ctx.fillRect(0, 0, ctx.canvas.width, ctx.canvas.height);
        pcs.push(ctx);
    }

    for (let ap = 0; ap < pc; ap++) {
        const ctx = pcs[ap];
        for (let as = 0; as < sc; as++) {
            ctx.strokeStyle = '#000000';
            ctx.lineWidth = 2;
            ctx.fillStyle = vis.colors_players[ap % vis.colors_players.length];
            ctx.beginPath();
            ctx.fillRect(0, 0, vis.canvas_player_width, 0.5 * poh);
            ctx.fillStyle = '#848484';
            if (cev.skillstatus[as] === 1) {
                ctx.strokeStyle = '#ff0000';
                ctx.lineWidth = 5;
            } else if (cev.skillstatus[as] === 2) {
                ctx.fillStyle = '#222222';
            }
            ctx.beginPath();
            ctx.rect((as + vis.canvas_player_skill_margin) * pow,
                vis.canvas_player_height - game.skills[as] * poh,
                (1 - 2 * vis.canvas_player_skill_margin) * pow,
                game.skills[as] * poh);
            ctx.stroke();
            ctx.fill();
            ctx.fillStyle = vis.colors_skills[as % vis.colors_skills.length];
            ctx.lineWidth = 2;
            ctx.beginPath();
            ctx.fillRect((as + vis.canvas_player_skill_margin) * pow,
                vis.canvas_player_height - cev.skills[ap][as] * poh,
                (1 - 2 * vis.canvas_player_skill_margin) * pow,
                cev.skills[ap][as] * poh);
            if (as === cev.actions[ap]) {
                ctx.fillStyle = '#ff0000';
                ctx.beginPath();
                ctx.fillRect((as + vis.canvas_player_skill_margin) * pow,
                    vis.canvas_player_height - cev.skills[ap][as] * poh,
                    (1 - 2 * vis.canvas_player_skill_margin) * pow,
                    pow);
            }
        }

        els.vis_player_scores_eval[ap].innerHTML = Math.round(cev.evalscore[ap] * 100) / 100;
        els.vis_player_scores_total[ap].innerHTML = Math.round(cev.score[ap] * 100) / 100;
    }

    // score
    const coh = vis.canvas_score_height / (vis.max_score + 2);
    const cow = vis.canvas_score_width / vis.total_rounds;
    const ctx = els.vis_score.getContext('2d');
    ctx.fillStyle = '#ffffff';
    ctx.beginPath();
    ctx.fillRect(0, 0, ctx.canvas.width, ctx.canvas.height);
    for(let ap = 0; ap < pc; ap++) {
        ctx.strokeStyle = vis.colors_players[ap % vis.colors_players.length];
        ctx.lineWidth = 2;
        ctx.beginPath();
        ctx.moveTo(0, vis.canvas_score_height - coh);
        let ax = 0;
        for(let ae = 0; ae <= vis.current_eval; ae++) {
            const end = ae === vis.current_eval ? vis.current_round + 1 : game.evals[ae].length;
            for(let ar = 0; ar < end; ar++) {
                ax++;
                ctx.lineTo(ax * cow, vis.canvas_score_height - (game.evals[ae][ar].score[ap] + 1) * coh);
            }
        }
        ctx.stroke();
        ax = 0;
        ctx.beginPath();
        ctx.strokeStyle = '#b7b7b7';
        ctx.lineWidth = 0.5;
        for (let ae = 0; ae < vis.current_eval; ae++) {
            ax += game.evals[ae].length;
            ctx.moveTo(ax * cow, 0);
            ctx.lineTo(ax * cow, vis.canvas_player_height);
        }
        ctx.stroke();
    }
}

function vis_toggle() {
    if (vis.playing) {
        vis_pause();
    } else {
        vis_play();
    }
}

function vis_play() {
    vis.playing = true;
    if (vis.animation !== -1) {
        window.clearInterval(vis.animation);
    }
    vis.animation = window.setInterval(vis_next, vis.timeout);
    els.nav_toggle.innerHTML = "Pause";
    vis.next_countdown = 0;
}

function vis_pause() {
    vis.playing = false;
    if (vis.animation !== -1) {
        window.clearInterval(vis.animation);
        vis.animation = -1;
    }
    els.nav_toggle.innerHTML = "Play";
}

function vis_next() {
    if (vis.playing && vis.current_round === game.evals[vis.current_eval].length - 1) {
        if (vis.next_countdown <= 0) {
            vis.next_countdown = vis.next_countdown_init;
        } else {
            vis.next_countdown--;
        }
        if (vis.next_countdown > 0) {
            return;
        }
    }
    vis.current_round++;
    if (vis.current_round >= game.evals[vis.current_eval].length) {
        vis.current_round = 0;
        vis.current_eval++;
        if (vis.current_eval >= game.evals.length) {
            vis.current_eval = 0;
        }
    }
    vis_apply_state();
}

function vis_prev() {
    vis.current_round--;
    if (vis.current_round < 0) {
        vis.current_eval--;
        if (vis.current_eval < 0) {
            vis.current_eval = game.evals.length - 1;
        }
        vis.current_round = game.evals[vis.current_eval].length - 1;
    }
    vis_apply_state();
}

function vis_timeout_changed() {
    const new_timeout = parseInt(els.nav_timeout.value);
    if (new_timeout !== vis.timeout) {
        vis.timeout = new_timeout;
        if (vis.playing) {
            vis_play();
        }
    }
}

function vis_eval_changed() {
    const new_eval = parseInt(els.nav_eval.value) - 1;
    if (new_eval !== vis.current_eval) {
        vis.current_eval = new_eval;
        if (new_eval >= game.evals.length) {
            vis.current_eval = game.evals.length - 1;
        } else if (new_eval < 0) {
            vis.current_eval = 0;
        }
        vis_apply_state();
    }
}

function vis_round_changed() {
    const new_round = parseInt(els.nav_round.value) - 1;
    if (new_round !== vis.current_round) {
        vis.current_round = new_round;
        if (new_round >= game.evals[vis.current_eval].length) {
            vis.current_round = game.evals[vis.current_eval].length - 1;
        } else if (new_round < 0) {
            vis.current_round = 0;
        }
        vis_apply_state();
    }
}

function load_log(log) {
    if (log.length < 10) {
        return;
    }
    els.log.value = log;
    game = {};
    game.skills = [];
    game.players = [];
    game.evals = [];

    game.ce = [];
    game.tp = 0;
    game.t = 0;
    const logsp = log.trim().replace("\r", "").split("\n");
    logsp.forEach(function (line) {
        const ls = line.split(" ");
        switch (ls[0]) {
            case 'GAME':
                game.playercount = parseInt(ls[1]);
                game.skillcount = parseInt(ls[2]);
                break;
            case 'SKILLS':
                game.skills = ls.slice(1).map(parseInt10);
                if (game.skills.length !== game.skillcount) {
                    console.warn(`Skills length does not match`);
                }
                break;
            case 'PNAME':
                game.players.push(ls.slice(1).join(' '));
                break;
            case 'PERMUTATION':
                game.permutation = ls.slice(1).map(parseInt10);
                break;
            case 'EVAL':
                game.evals.push([]);
                game.ce = game.evals[game.evals.length - 1];
                break;
            case 'TARGET':
                game.tp = parseInt(ls[1]);
                game.t = parseInt(ls[2]);
                break;
            case 'ACTION':
                game.ce.push({
                    target: game.t,
                    target_player: game.tp,
                    actions: ls.slice(1).map(parseInt10)
                });
                if (ls.length - 1 !== game.playercount) {
                    console.warn(`Actions length does not match`);
                }
                break;
            case 'RESET':
                break;
            case 'RESULTS':
                break;
            default:
                if (game.players.indexOf(line.split(' : ')[0]) === -1) {
                    console.warn(`Invalid log line ${line}`);
                }
        }
    });
    delete game.ce;
    delete game.t;
    delete game.tp;
    if (game.players.length !== game.playercount) {
        console.warn(`Players length does not match`);
    }

    const pc = game.playercount;
    const sc = game.skillcount;
    game.sc = [];
    for (let i = 0; i < pc; i++) {
        game.sc.push(0);
    }
    game.evals.forEach(function (eval) {
        const esc = [];
        const sk = [];
        const ss = [];
        for (let i = 0; i < pc; i++) {
            esc.push(0);
            let ask = [];
            for (let j = 0; j < sc; j++) {
                ask.push(0);
            }
            sk.push(ask);
        }
        for (let i = 0; i < sc; i++) {
            ss.push(0);
        }
        for (let i = 0; i < eval.length; i++) {
            const av = eval[i];
            if (ss[av.target] !== 1) {
                for (let j = 0; j < sc; j++) {
                    if (ss[j] === 1) {
                        ss[j] = 2;
                    }
                }
                ss[av.target] = 1;
            }
            let cnt = 0;
            for (let j = 0; j < pc; j++) {
                sk[j][av.actions[j]]++;
                if (sk[j][av.target] === game.skills[av.target]) {
                    cnt++;
                }
            }
            if (cnt) {
                for (let j = 0; j < pc; j++) {
                    if (sk[j][av.target] === game.skills[av.target]) {
                        const score = 1 / cnt;
                        esc[j] += score;
                        game.sc[j] += score;
                    }
                }
            }
            av.skills = sk.slice().map(function (x) {
                return x.slice();
            });
            av.skillstatus = ss.slice();
            av.evalscore = esc.slice();
            av.score = game.sc.slice();
        }
    });
    game.endscore = game.sc.slice();
    delete game.sc;

    vis_init();
}

function load_file(file) {
    if (!file) {
        els.log.value = 'Log';
        return;
    }
    if (predef_games[file]) {
        load_log(predef_games[file]);
        return;
    }
    const script = document.createElement('script');
    script.src = `${file}.js`;
    script.onload = function () {
        load_log(predef_games[file]);
    };
    els.jsloader.appendChild(script);
}
